import numpy as np
from autoeap.detrender import detrend_wrt_PDM
from numpy.testing import assert_array_almost_equal

import os
PACKAGEDIR = os.path.abspath(os.path.dirname(__file__))

def test_detrender():
    time,flux,flux_err = np.genfromtxt(os.path.join(PACKAGEDIR,"EPIC220198696_c8_autoEAP.lc"),skip_header=1,unpack=True)

    corr_flux = detrend_wrt_PDM(time,flux,flux_err)

    corr_flux_orig = np.genfromtxt(os.path.join(PACKAGEDIR,"EPIC220198696_c8_autoEAP.lc_corr"))

    assert_array_almost_equal(corr_flux_orig,corr_flux)



import numpy as np
import autoeap
from numpy.testing import assert_array_almost_equal

import os
PACKAGEDIR = os.path.abspath(os.path.dirname(__file__))

def test_raw_lightcurve():
    time,flux,flux_err = autoeap.createlightcurve('EPIC220198696',campaign=8)

    lc = np.genfromtxt(os.path.join(PACKAGEDIR,"EPIC220198696_c8_autoEAP.lc"),skip_header=1).T

    assert_array_almost_equal(time,lc[0])
    assert_array_almost_equal(flux,lc[1].astype(np.float32))
    assert_array_almost_equal(flux_err,lc[2].astype(np.float32))



# Purpose: Cython wrapper for the PDM module
# Author: Attila Bódi
# Version: 0.1  2021APR01

import cython
import numpy as np
cimport numpy as np

# declare the interface to the C code
cdef extern void pdm_binned_step(
        double* t,
        double* y,
        double* w,
        double* freqs,
        double* power,
        const int ndata,
        const int nfreqs,
        const float dphi,
        const int nbins)

cdef extern void pdm_binned_linterp(
        double* t,
        double* y,
        double* w,
        double* freqs,
        double* power,
        const int ndata,
        const int nfreqs,
        const float dphi,
        const int nbins)

cdef extern void pdm_binless_tophat(
        double* t,
        double* y,
        double* w,
        double* freqs,
        double* power,
        const int ndata,
        const int nfreqs,
        const float dphi,
        const int nbins)

cdef extern void pdm_binless_gauss(
        double* t,
        double* y,
        double* w,
        double* freqs,
        double* power,
        const int ndata,
        const int nfreqs,
        const float dphi,
        const int nbins)

@cython.boundscheck(False)
@cython.wraparound(False)
def PDM(    np.ndarray[np.double_t, ndim=1, mode="c"] t not None,
            np.ndarray[np.double_t, ndim=1, mode="c"] y not None,
            np.ndarray[np.double_t, ndim=1, mode="c"] w not None,
            np.ndarray[np.double_t, ndim=1, mode="c"] freqs not None,
            float dphi=0.05,
            int nbins=10,
            str kind = 'binned_linterp',

    ):
    """
    ``Polyfit`` fits polynomial chain to phase curve

    Parameters
    ----------
    t: array
        time values5
    y: array
        Flux/mag values
    w: array
        weights
    freqs: array
        freqs
    """

    cdef int ndata  = t.size
    cdef int nfreqs = freqs.size
    cdef np.ndarray[np.float_t, ndim=1, mode="c"] power   = np.zeros((freqs.size),dtype=np.floating)

    # --- Call C funtion ---
    if kind == 'binned_step':
        pdm_binned_step(
                &t[0],
                &y[0],
                &w[0],
                &freqs[0],
                &power[0],
                ndata,
                nfreqs,
                dphi,
                nbins)
    elif kind == 'binned_linterp':
        pdm_binned_linterp(
                &t[0],
                &y[0],
                &w[0],
                &freqs[0],
                &power[0],
                ndata,
                nfreqs,
                dphi,
                nbins)
    elif kind == 'binless_tophat':
        pdm_binless_tophat(
                &t[0],
                &y[0],
                &w[0],
                &freqs[0],
                &power[0],
                ndata,
                nfreqs,
                dphi,
                nbins)
    elif kind == 'binless_gauss':
        pdm_binless_gauss(
                &t[0],
                &y[0],
                &w[0],
                &freqs[0],
                &power[0],
                ndata,
                nfreqs,
                dphi,
                nbins)
    else:
        raise KeyError('Function not available. Please use one of the followings: ' + \
                            'binless_tophat, binless_gauss, binned_linterp, binned_step')

    return power



# These functions are from lightkurve 1.x
# Source: https://github.com/lightkurve/lightkurve/blob/v1.11.x/lightkurve/targetpixelfile.py

from astropy.io import fits
import lightkurve
import os
import numpy as np

__all__ = ['save_aperture_fits']

def _boolean_mask_to_bitmask(aperture_mask):
    """Takes in an aperture_mask and returns a Kepler-style bitmask
    Parameters
    ----------
    aperture_mask : array-like
        2D aperture mask. The mask can be either a boolean mask or an integer
        mask mimicking the Kepler/TESS convention; boolean or boolean-like masks
        are converted to the Kepler/TESS conventions.  Kepler bitmasks are
        returned unchanged except for possible datatype conversion.
    Returns
    -------
    bitmask : numpy uint8 array
        A bitmask incompletely mimicking the Kepler/TESS convention: Bit 2,
        value = 3, means "pixel was part of the custom aperture".  The other
        bits have no meaning and are currently assigned a value of 1.
    """
    # Masks can either be boolean input or Kepler pipeline style
    clean_mask = np.nan_to_num(aperture_mask)

    contains_bit2 = (clean_mask.astype(np.int) & 2).any()
    all_zeros_or_ones = ( (clean_mask.dtype in ['float', 'int']) &
                            ((set(np.unique(clean_mask)) - {0,1}) == set()) )
    is_bool_mask = ( (aperture_mask.dtype == 'bool') | all_zeros_or_ones )

    if is_bool_mask:
        out_mask = np.ones(aperture_mask.shape, dtype=np.uint8)
        out_mask[aperture_mask == 1] = 3
        out_mask = out_mask.astype(np.uint8)
    elif contains_bit2:
        out_mask = aperture_mask.astype(np.uint8)
    else:
        out_mask = None
    return out_mask

def _header_template(extension):
    """Returns a template `fits.Header` object for a given extension."""
    template_fn = os.path.join(lightkurve.__path__[0], "data",
                               "tpf-ext{}-header.txt".format(extension))
    return fits.Header.fromtextfile(template_fn)


def save_aperture_fits(aperture_mask,filename,overwrite=False):
    """Saves an `ImageHDU` object containing the 'APERTURE' extension."""
    bitmask = _boolean_mask_to_bitmask(aperture_mask)
    hdu = fits.ImageHDU(bitmask)

    # Set the header from the template TPF again
    template = _header_template(2)
    for kw in template:
        if kw not in ['XTENSION', 'NAXIS1', 'NAXIS2', 'CHECKSUM', 'BITPIX']:
            hdu.header[kw] = (template[kw],
                              template.comments[kw])

    # Override the defaults where necessary
    for keyword in ['CTYPE1', 'CTYPE2', 'CRPIX1', 'CRPIX2', 'CRVAL1', 'CRVAL2', 'CUNIT1',
                    'CUNIT2', 'CDELT1', 'CDELT2', 'PC1_1', 'PC1_2', 'PC2_1', 'PC2_2']:
            hdu.header[keyword] = ""  # override wcs keywords
    hdu.header['EXTNAME'] = 'APERTURE'

    hdu.writeto(filename,overwrite=overwrite)

    pass



from os import path
import sys
from setuptools import setup
from distutils.extension import Extension
from Cython.Build import cythonize

import numpy

sys.path.insert(0, "src")
from version import __version__

# Load requirements
requirements = None
with open('requirements.txt') as file:
    requirements = file.read().splitlines()

# If Python3: Add "README.md" to setup.
# Useful for PyPI. Irrelevant for users using Python2.
try:
    this_directory = path.abspath(path.dirname(__file__))
    with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
        long_description = f.read()
except:
    long_description = ' '

# Command-line tools
entry_points = {'console_scripts': [
    'autoeap = autoeap:autoeap_from_commandline'
]}

extensions = Extension("PDM",
                         sources=["src/PDM/PDM_pyc.pyx", "src/PDM/pdm.c"],
                         include_dirs=[numpy.get_include(),'/opt/local/include'],
                         library_dirs=['/opt/local/lib'],
                         libraries=['m'],
                         define_macros=[('NPY_NO_DEPRECATED_API', 'NPY_1_7_API_VERSION')])

desc='Automated version of Extended Aperture Photometry developed for high amplitude K2 variable stars.'

setup(name='autoeap',
      version=__version__,
      description=desc,
      long_description=long_description,
      long_description_content_type='text/markdown',
      author='Attila Bodi & Pal Szabo',
      author_email='bodi.attila@csfk.org',
      url='https://github.com/konkolyseismolab/autoeap/',
      package_dir={'autoeap':'src'},
      packages=['autoeap'],
      install_requires=requirements,
      entry_points=entry_points,
      ext_modules = cythonize(extensions, compiler_directives={'language_level': 3})
     )



import PDM
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from astropy.timeseries import LombScargle
from scipy.optimize import minimize
from wotan import flatten
from astropy.stats import sigma_clip
import sys, os
import warnings

__all__ = ['detrend_wrt_PDM']

def weights(err):
    """ generate observation weights from uncertainties """
    w = np.power(err, -2)
    return w/sum(w)

def theta_by_coeffs(coeffs,*params):
    x,y,yerr,testf,const = params

    ydetrended = y.copy()
    for order,coeff in enumerate(coeffs[::-1]):
        ydetrended -= coeff*x**(order+1)
    ydetrended -= const

    w = weights( yerr )

    kind='binned_linterp'

    with warnings.catch_warnings(record=True) as warnmessage:
        testf = np.atleast_1d(testf)
        x =          np.asarray(x,dtype=np.floating)
        ydetrended = np.asarray(ydetrended,dtype=np.floating)
        testf =      np.asarray(testf,dtype=np.floating)
        w =          np.asarray(w,dtype=np.floating)
        pows = PDM.PDM(x, ydetrended, w, testf, kind=kind, nbins=50, dphi=0.05)

    return 1 - pows


def get_trend(x,y,err,freq,polyorder=5):
    p0 = np.polyfit(x,y,polyorder)
    const = p0[-1]

    # Set bounds to first and second orders to avoid divergence
    bounds = tuple([(-np.inf,np.inf) for _ in range(len(p0)-3)])
    try: bounds += ((p0[-3] - np.abs(p0[-3])*2., p0[-3] + np.abs(p0[-3])*2. ),)
    except IndexError: pass
    try: bounds += ((p0[-2] - np.abs(p0[-2])*1., p0[-2] + np.abs(p0[-2])*1. ),)
    except IndexError: pass

    if len(bounds) == 0: bounds=None

    res = minimize(theta_by_coeffs, p0[:-1], args=(x,y,err,freq,const), method='Powell', bounds=bounds)

    return res,const

def clean_lightcurve(x,y,sigma=4,plotting=False):
    class HiddenPrints:
        def __enter__(self):
            self._original_stdout = sys.stdout
            sys.stdout = open(os.devnull, 'w')

        def __exit__(self, exc_type, exc_val, exc_tb):
            sys.stdout.close()
            sys.stdout = self._original_stdout

    goodpts = np.isfinite(y)
    freq,power = LombScargle(x[goodpts], y[goodpts]).autopower( normalization='psd',
                                                                nyquist_factor=0.8,
                                                                minimum_frequency=1/x.ptp(),
                                                                samples_per_peak=50)
    bestper = 1/freq[np.argmax(power)]

    with HiddenPrints():
        # Fit detrended data with cosine to remove outliers
        flatten_lc, cos_trend = flatten(x, y,
                                    window_length=bestper*0.5,
                                    return_trend=True,
                                    method='median')

    # Handle if trend is all nan
    if np.all(np.isnan(cos_trend)):
        flatten_lc = y
        cos_trend  = np.ones_like(cos_trend) * np.mean(y)

    # Find outliers after removing cosine fit
    goodpts = sigma_clip(flatten_lc, sigma_lower=sigma, sigma_upper=sigma, maxiters=5, cenfunc='mean', stdfunc='std')
    goodpts = ~goodpts.mask

    if plotting:
        fig = plt.figure(figsize=(14,5))
        plt.title('Original & w/out outliers')
        plt.plot(x[goodpts], y[goodpts],'k.', ms=2)
        plt.plot(x[~goodpts], y[~goodpts],'rx', ms=6)
        plt.plot(x,cos_trend,c='C1')
        plt.show()
        plt.close()

    return  goodpts


def detrend_wrt_PDM(time,flux,fluxerr,polyorder='auto',sigma=10,show_plots=False,save_plots=False,filename=None,debug=False):
    """
    Deterending with a polynomial that is fitted w.r.t. phase dispersion minimization.

    Parameters
    ----------
    time : array-like
        Time values.
    flux : array-like
        Corresponding flux values.
    fluxerr : array-like
        Corresponding flux errors.
    polyorder : int , default : 9
        The order of the detrending polynomial.
    show_plots: bool, default: False
        If `True` the plot will be displayed.
    save_plots: bool, default: False
        If `True` the plot will be saved to a subdirectory.
    filename : string
        If `save_plots` is `True`, the name of the output PNG file,
        without extension.
    debug : bool, default: False
        If `True` debug plots/prints will be displayed.

    Returns
    -------
    corrflux : array-like
        Detrended flux values.
    """

    with warnings.catch_warnings(record=True) as w:
        x = np.ascontiguousarray(time, dtype=np.float64)
        yorig = np.ascontiguousarray(flux, dtype=np.float64)
        err = np.ascontiguousarray(fluxerr, dtype=np.float64)
    y = yorig.copy()
    xorig = x.copy()
    errorig = err.copy()

    goodpts = np.isfinite(y)
    x = x[goodpts]
    y = y[goodpts]
    err = err[goodpts]

    # flatten cannot handle negative values
    if y.mean() <=0:
        y -= y.min()
        yorig -= yorig.min()

    ## Clean light curve roughly by rolling median filtering in time and sigma clipping

    goodpts = clean_lightcurve(x,y,sigma=sigma,plotting=debug)

    # Keep only good points
    x = x[goodpts]
    y = y[goodpts]
    err = err[goodpts]

    # Get period
    lsfreqs,lspow = LombScargle(x,y).autopower(nyquist_factor=1,samples_per_peak=50,minimum_frequency=0.05)
    testf = lsfreqs[np.argmax(lspow)]

    # If test period is larger than data length, remove linear
    if testf <= 1/np.ptp(x):
        y -= np.poly1d(np.polyfit(x,y,1))(x)

        lsfreqs,lspow = LombScargle(x,y).autopower(normalization='psd',
                                                nyquist_factor=1,
                                                minimum_frequency=0.05,
                                                samples_per_peak=50)

        testf = lsfreqs[np.argmax(lspow)]

        # If period is still larger than data length
        if testf <= 1/np.ptp(x):
            testf = 1/np.ptp(x)

    del lsfreqs,lspow

    if debug:
        print('Best period is', round(1/testf,8),'days' )

        fig,ax = plt.subplots(1,2,figsize=(14,5))
        ax[1].plot( (x*testf)%1, y, '.', ms=2)
        ax[0].plot( x, y, '.', ms=2)
        plt.show()
        plt.close()

    #### Normalize light curve
    meanx = np.nanmean(x)
    meany = np.nanmean(y)
    x -= meanx
    y -= meany

    # Estimate polyorder based on covered cycles
    if polyorder == 'auto':
        if x.ptp()/(1/testf) < 5:
            polyorder = 0
        elif x.ptp()/(1/testf) < 200:
            polyorder = 5
        elif x.ptp()/(1/testf) < 500:
            polyorder = 7
        else:
            polyorder = 9

    # Run PDM-based polynomial fitting
    res,const = get_trend(x,y,err,freq=testf,polyorder=polyorder)

    # If PDM-based trend fit fails use simple polyfit instead
    if np.std( y-np.poly1d(np.append(res.x,const))(x) )/np.std(y) > 1.0e3:
        class result:
            def __init__(self,x):
                self.x = x

        _pol = np.polyfit(x,y,polyorder)
        res = result(_pol[:-1])
        const = _pol[-1]

        del _pol
        del result

    ### Plot detrending

    gs = GridSpec(3,2)

    fig = plt.figure(figsize=(15,10))
    ax1 = plt.subplot(gs[0,:])
    ax2 = plt.subplot(gs[1,:])
    ax3 = plt.subplot(gs[2,0])
    ax4 = plt.subplot(gs[2,1])

    ax1.set_title('Black: inital polinomial -- Red: final polynomial')
    ax1.scatter(x+meanx-2454833,y+meany,s=4,c=x,cmap='copper')
    ax1.plot(x+meanx-2454833,  np.poly1d(np.polyfit(x,y,polyorder))(x) +meany,'k')

    ax1.plot(x+meanx-2454833, np.poly1d(np.append(res.x,const))(x) +meany,'r' )
    ax1.grid()
    ax1.set_ylabel('Flux')

    ax2.set_title('Corrected light curve')
    ax2.scatter(x+meanx-2454833, y-np.poly1d(np.append(res.x,const))(x) +meany,s=4,c=x,cmap='copper'  )
    ax2.grid()
    ax2.set_ylabel('Flux')
    ax2.set_xlabel('Time - 2454833 [BKJD days]')

    ax3.set_title('Original')
    ax3.scatter( (x*testf)%1, y+meany, c=x, s=4,cmap='copper')
    ax3.grid()
    ax3.set_xlabel('Phase')
    ax3.set_ylabel('Flux')
    ax4.set_title('Corrected')
    ax4.scatter( (x*testf)%1, y-np.poly1d(np.append(res.x,const))(x) +meany, c=x, s=4,cmap='copper')
    ax4.grid()
    ax4.set_xlabel('Phase')

    plt.tight_layout()
    if save_plots and filename is not None: plt.savefig(filename+'.png',format='png',dpi=80)
    if show_plots: plt.show()
    plt.close()

    ### Update variables

    ycorr = yorig - np.poly1d(np.append(res.x,const))(xorig-meanx)

    ### Return detrended light curve

    return ycorr
